/*****************************************************************************
 * CryptEng.c
 *
 * SAFER SK-128 Algorithm
 * Converted from J.L. Massey 9-9-95 document - Pascal example.
 *
 * {12/10/96,TAC}
 *
 * Choosing both halves of the key to coincide gives the same result as
 * for encrypting with this same 64 bit sequence as the key for SAFER
 * SK-64.
 *
 ****************************************************************************/

#include <string.h>

#define ROUNDS 12

char dbuff[512];

typedef unsigned char  uchar;
typedef unsigned short ushort;
typedef unsigned int   uint;
typedef unsigned long  ulong;

/****************************************************************************/
uchar  a1, a2, a3, a4, a5, a6, a7, a8; 
uchar  b1, b2, b3, b4, b5, b6, b7, b8;
uchar  r;
uchar  k[25][8], ka[9], kb[9];
ushort logtab[256], exptab[256], i, j;

/****************************************************************************/
__inline void mat1(uchar *a1, uchar *a2, uchar *b1, uchar *b2)
{ 
  *b2 = *a1 + *a2; 
  *b1 = *b2 + *a1; 
}

/****************************************************************************/
__inline void invmat1(uchar *a1, uchar *a2, uchar *b1, uchar *b2)
{
  *b1 =  *a1 - *a2; 
  *b2 = -*b1 + *a2; 
}

/*****************************************************************************
 *
 *    The program here computes the powers of the primitive element 45 of
 * the finite field GF(257) and stores these in the table "exptab".
 * Corresponding logarithms to the base 45 are stored in the table "logtab".
 *
 */
static void BuildKeyTables()
{

  logtab[1] = 0; 
  exptab[0] = 1;
  
  for(i=1; i<256; i++)
  {
    exptab[i] = (45 * exptab[i - 1]) % 257;

    if(exptab[i] < 256)  
      logtab[exptab[i]] = i;
  }
  
  exptab[128] = 0; 
  logtab[0] = 128;

}

/*****************************************************************************
 *
 *   Decrypt InBuff using Key. BuffSize Mod 8 is assumed to be 0.
 *
 */
void Decrypt(const uchar *InBuff, uchar *OutBuff, ushort *Key, uint BuffSize)
{
  int index;
  uint BytesProcessed=0;

  BuildKeyTables();

  r = ROUNDS;

  for(i=0; i<8; i++)
  {
    ka[i] = (uchar)Key[i];
    kb[i] = (uchar)Key[i+8];
  }

  /* The next instruction app}s a "parity byte" to keys Ka and Kb. */
  ka[8] = ka[0] ^ ka[1] ^ ka[2] ^ ka[3] ^ ka[4] ^ ka[5] ^ ka[6] ^ ka[7];
  kb[8] = kb[0] ^ kb[1] ^ kb[2] ^ kb[3] ^ kb[4] ^ kb[5] ^ kb[6] ^ kb[7];
  
  /*
   *  The next instructions implement the key schedule
   *  needed to derive keys K1, K2, ... K2r+1 from the
   *  128 bit input key (Ka, Kb).
   */

  /* K0 is set equal to Kb. */
  for(j=0; j<8; j++) 
    k[0][j] = kb[j];

  /* Each byte of the key Ka is right rotated by 3. */
  for(j=0; j<9; j++) 
    ka[j] = (ka[j] >> 3) + (ka[j] << 5);

  for(i=0; i<r; i++)
  {
    /* Each byte of keys Ka and Kb is further left rotated by 6. */
    for(j=0; j<9; j++)
    {
      ka[j] = (ka[j] << 6) + (ka[j] >> 2);
      kb[j] = (kb[j] << 6) + (kb[j] >> 2);
    }

    /* The key biases are added to give the keys K2i+1 and K2i+2. */
    for(j=0;j<8;j++)
    {
      k[2*i+1][j] = ka[(((j+1)+2*(i+1)-2) % 9)] + 
                       exptab[exptab[18*(i+1)+(j+1)]];
      k[2*i+2][j] = kb[(((j+1)+2*(i+1)-1) % 9)] + 
                       exptab[exptab[18*(i+1)+9+(j+1)]];
    }

  }/* for i */

  do
  {
    a1 = InBuff[BytesProcessed+0];
    a2 = InBuff[BytesProcessed+1];
    a3 = InBuff[BytesProcessed+2];
    a4 = InBuff[BytesProcessed+3];
    a5 = InBuff[BytesProcessed+4];
    a6 = InBuff[BytesProcessed+5];
    a7 = InBuff[BytesProcessed+6];
    a8 = InBuff[BytesProcessed+7];


    /* Key 2r is now mixed bit and byte added to produce the cryptogram. */
    index = r<<1;

    a1 = a1 ^ k[index][0]; 
    a2 = a2 - k[index][1];
    a3 = a3 - k[index][2]; 
    a4 = a4 ^ k[index][3];
    a5 = a5 ^ k[index][4]; 
    a6 = a6 - k[index][5];
    a7 = a7 - k[index][6]; 
    a8 = a8 ^ k[index][7]; 

    /* The r rounds of encryption here. */
    for(i=0; i<r; i++)
    {

      /* The result now enters the first level of the linear layer. */
      invmat1(&a1, &a2, &b1, &b2); 
      invmat1(&a3, &a4, &b3, &b4);
      invmat1(&a5, &a6, &b5, &b6); 
      invmat1(&a7, &a8, &b7, &b8);

      /* The result now enters the second level of the linear layer. */
      invmat1(&b1, &b5, &a1, &a2); 
      invmat1(&b2, &b6, &a3, &a4);
      invmat1(&b3, &b7, &a5, &a6); 
      invmat1(&b4, &b8, &a7, &a8);

      /* The result now enters the third level of the linear layer. */
      invmat1(&a1, &a5, &b1, &b2); 
      invmat1(&a2, &a6, &b3, &b4);
      invmat1(&a3, &a7, &b5, &b6); 
      invmat1(&a4, &a8, &b7, &b8);

      /* Key 2(r-i)-1 is now mixed byte and bit added to the result. */
      index = 2*(r-i)-1;

      b1 = b1 - k[index][0]; 
      b2 = b2 ^ k[index][1];
      b3 = b3 ^ k[index][2]; 
      b4 = b4 - k[index][3];
      b5 = b5 - k[index][4]; 
      b6 = b6 ^ k[index][5];
      b7 = b7 ^ k[index][6]; 
      b8 = b8 - k[index][7];

      /* The result now passes through the nonlinear layer. */
      a1 = logtab[b1]; 
      a2 = exptab[b2]; 
      a3 = exptab[b3]; 
      a4 = logtab[b4];
      a5 = logtab[b5]; 
      a6 = exptab[b6]; 
      a7 = exptab[b7]; 
      a8 = logtab[b8];

      /* Key 2(r-i-1) is mixed bit and byte added to the round input. */
      index = 2*(r-i-1);

      a1 = a1 ^ k[index][0]; 
      a2 = a2 - k[index][1];
      a3 = a3 - k[index][2]; 
      a4 = a4 ^ k[index][3];
      a5 = a5 ^ k[index][4]; 
      a6 = a6 - k[index][5];
      a7 = a7 - k[index][6]; 
      a8 = a8 ^ k[index][7];
    }

    /* Feedback Cypher */
    if(BytesProcessed)/* P'[BlockNum] = P[BlockNum] ^ C[BlockNum-1] */
    {
      a1 ^= InBuff[BytesProcessed-8];
      a2 ^= InBuff[BytesProcessed-7];
      a3 ^= InBuff[BytesProcessed-6];
      a4 ^= InBuff[BytesProcessed-5];
      a5 ^= InBuff[BytesProcessed-4];
      a6 ^= InBuff[BytesProcessed-3];
      a7 ^= InBuff[BytesProcessed-2];
      a8 ^= InBuff[BytesProcessed-1];
    }
    
    OutBuff[BytesProcessed+0] = (ushort)a1;
    OutBuff[BytesProcessed+1] = (ushort)a2;
    OutBuff[BytesProcessed+2] = (ushort)a3;
    OutBuff[BytesProcessed+3] = (ushort)a4;
    OutBuff[BytesProcessed+4] = (ushort)a5;
    OutBuff[BytesProcessed+5] = (ushort)a6;
    OutBuff[BytesProcessed+6] = (ushort)a7;
    OutBuff[BytesProcessed+7] = (ushort)a8;

    BytesProcessed += 8;
  }
  while(BytesProcessed < BuffSize);
}

/*****************************************************************************
 *
 *   The program here computes the powers of the primitive element 45 of
 * the finite field GF(257) and stores these in the table "exptab".
 * Corresponding logarithms to the base 45 are stored in the table "logtab".
 *
 */
void Encrypt(const uchar *InBuff, uchar *OutBuff, ushort *Key, uint BuffSize)
{
  int index;
  uint BytesProcessed=0;

  BuildKeyTables();
  r = ROUNDS;

  for(i=0; i<8; i++)
  {
    ka[i] = (uchar)Key[i];
    kb[i] = (uchar)Key[i+8];
  }

  /* The next instruction app}s a "parity byte" to keys Ka and Kb. */
  ka[8] = ka[0] ^ ka[1] ^ ka[2] ^ ka[3] ^ ka[4] ^ ka[5] ^ ka[6] ^ ka[7];
  kb[8] = kb[0] ^ kb[1] ^ kb[2] ^ kb[3] ^ kb[4] ^ kb[5] ^ kb[6] ^ kb[7];
  
  /* 
   *  The next instructions implement the key schedule
   *  needed to derive keys K1, K2, ... K2r+1 from the
   *  128 bit input key (Ka, Kb).
   */

  /* K0 is set equal to Kb. */
  for(j=0; j<8; j++) 
    k[0][j] = kb[j];

  /* Each byte of the key Ka is right rotated by 3.} */
  for(j=0; j<9; j++) 
    ka[j] = (ka[j] >> 3) + (ka[j] << 5);

  for(i=0; i<r; i++)
  {
    /* Each byte of keys Ka and Kb is further left rotated by 6. */
    for(j=0; j<9; j++)
    {
      ka[j] = (ka[j] << 6) + (ka[j] >> 2);
      kb[j] = (kb[j] << 6) + (kb[j] >> 2);
    }

    /* The key biases are added to give the keys K2i+1 and K2i+2. */
    for(j=0;j<8;j++)
    {
      k[2*i+1][j] = ka[(((j+1)+2*(i+1)-2) % 9)] + 
                       exptab[exptab[18*(i+1)+(j+1)]];
      k[2*i+2][j] = kb[(((j+1)+2*(i+1)-1) % 9)] + 
                       exptab[exptab[18*(i+1)+9+(j+1)]];
    }
  }

  do
  {
    a1 = InBuff[BytesProcessed+0];
    a2 = InBuff[BytesProcessed+1];
    a3 = InBuff[BytesProcessed+2];
    a4 = InBuff[BytesProcessed+3];
    a5 = InBuff[BytesProcessed+4];
    a6 = InBuff[BytesProcessed+5];
    a7 = InBuff[BytesProcessed+6];
    a8 = InBuff[BytesProcessed+7];

    /* Feedback Cypher */
    if(BytesProcessed)/* P'[BlockNum] = P[BlockNum] ^ C[BlockNum-1] */
    {
      a1 ^= OutBuff[BytesProcessed-8];
      a2 ^= OutBuff[BytesProcessed-7];
      a3 ^= OutBuff[BytesProcessed-6];
      a4 ^= OutBuff[BytesProcessed-5];
      a5 ^= OutBuff[BytesProcessed-4];
      a6 ^= OutBuff[BytesProcessed-3];
      a7 ^= OutBuff[BytesProcessed-2];
      a8 ^= OutBuff[BytesProcessed-1];
    }
    
    /* The r rounds of encryption here. */
    for(i=0; i<r; i++)
    {
      /* Key 2i is mixed bit and byte added to the round input. */
      index = i<<1;

      a1 = a1 ^ k[index][0]; 
      a2 = a2 + k[index][1];
      a3 = a3 + k[index][2]; 
      a4 = a4 ^ k[index][3];
      a5 = a5 ^ k[index][4]; 
      a6 = a6 + k[index][5];
      a7 = a7 + k[index][6]; 
      a8 = a8 ^ k[index][7];

      /* The result now passes through the nonlinear layer. */
      b1 = exptab[a1]; 
      b2 = logtab[a2]; 
      b3 = logtab[a3]; 
      b4 = exptab[a4];
      b5 = exptab[a5]; 
      b6 = logtab[a6]; 
      b7 = logtab[a7]; 
      b8 = exptab[a8];

      /* Key 2i+1 is now mixed byte and bit added to the result. */
      index = (i<<1)+1;

      b1 = b1 + k[index][0]; 
      b2 = b2 ^ k[index][1];
      b3 = b3 ^ k[index][2]; 
      b4 = b4 + k[index][3];
      b5 = b5 + k[index][4]; 
      b6 = b6 ^ k[index][5];
      b7 = b7 ^ k[index][6]; 
      b8 = b8 + k[index][7];

      /* The result now enters the first level of the linear layer. */
      mat1(&b1, &b2, &a1, &a2); 
      mat1(&b3, &b4, &a3, &a4);
      mat1(&b5, &b6, &a5, &a6); 
      mat1(&b7, &b8, &a7, &a8);

      /* The result now enters the second level of the linear layer. */
      mat1(&a1, &a3, &b1, &b2); 
      mat1(&a5, &a7, &b3, &b4);
      mat1(&a2, &a4, &b5, &b6); 
      mat1(&a6, &a8, &b7, &b8);

      /* The result now enters the third level of the linear layer. */
      mat1(&b1, &b3, &a1, &a2); 
      mat1(&b5, &b7, &a3, &a4);
      mat1(&b2, &b4, &a5, &a6); 
      mat1(&b6, &b8, &a7, &a8);
    }/* for i */

    /* Key 2r is now mixed bit and byte added to produce the cryptogram. */
    index = r<<1;

    a1 = a1 ^ k[index][0]; 
    a2 = a2 + k[index][1];
    a3 = a3 + k[index][2]; 
    a4 = a4 ^ k[index][3];
    a5 = a5 ^ k[index][4]; 
    a6 = a6 + k[index][5];
    a7 = a7 + k[index][6]; 
    a8 = a8 ^ k[index][7];

    OutBuff[BytesProcessed+0] = (ushort)a1;
    OutBuff[BytesProcessed+1] = (ushort)a2;
    OutBuff[BytesProcessed+2] = (ushort)a3;
    OutBuff[BytesProcessed+3] = (ushort)a4;
    OutBuff[BytesProcessed+4] = (ushort)a5;
    OutBuff[BytesProcessed+5] = (ushort)a6;
    OutBuff[BytesProcessed+6] = (ushort)a7;
    OutBuff[BytesProcessed+7] = (ushort)a8;

    BytesProcessed += 8;
  }
  while(BytesProcessed < BuffSize);
}

