/***************************************************************************
 *
 *                 CiBApi.c (Micro-In-A-Can API source file) 
 *
 *                           { Ver 1.5 03/02/98 }
 *
 ***************************************************************************/

#include <memory.h>
#include <stdlib.h>

#include "cibapi.h"

WORD   LAST_ERROR      = CMD_OK;
WORD   ExecTimeMS      = 0;
WORD   MinTimeMS       = 0;
BYTE   CiBsFound       = 0;
BOOL   SearchParamsSet = FALSE;
IBADDR CurrCiB;
IBHOST SearchList;
BYTE   CiBROMs[MAX_CIB_NUM << 3]; 
BYTE   CiBPorts[MAX_CIB_NUM];

#define VERSION_MAJOR          1
#define VERSION_MINOR          500

/**************************************************************************
 * Returns the API version. Version major in LS WORD and minor in MS WORD
 */
DWORD DLLEXPORT GetCiBAPIVersion()
{
  return (((DWORD)VERSION_MINOR)<<16) + VERSION_MAJOR;
}
/****************************************************************************
 *
 *   Releases search list memory.
 *
 */
static void ReleaseCiBSearchList()
{
  BYTE i;

  if(SearchParamsSet)
  {
    SearchParamsSet = FALSE;

    for(i = 0; i < SearchList.AdapterCount; i++)
      free(SearchList.lpAdapter[i]);
  }
}
/**************************************************************************
 *
 *   Set parameters for CiB search. If SrchSer is TRUE copy serial port
 * device strings to search parameter structure.
 *
 */
void DLLEXPORT SetCiBSearchParams(BYTE SPNum, char *DevName[])
{
   BYTE i, n;

  if(SearchParamsSet)
    ReleaseCiBSearchList();

   SearchParamsSet = TRUE;
   SearchList.AdapterCount = MAX_PARALLEL_PORTS + SPNum;

   /* Initialize parallel port info */
   for (i = 0; i < MAX_PARALLEL_PORTS; i++)
   {
      SearchList.lpAdapter[i] =
         (LPADAPTERINFO) malloc(sizeof(ADAPTERINFO));

      SearchList.lpAdapter[i]->Type         = PARALLEL_DS1481;
      SearchList.lpAdapter[i]->Num          = i+1;
      SearchList.lpAdapter[i]->DevString[0] = 0;
   }

   n = MAX_PARALLEL_PORTS;
   /* Initialize serial port info */
   for (i = 0; i < SPNum; i++)
   {
      SearchList.lpAdapter[i+n] =
         (LPADAPTERINFO) malloc(sizeof(ADAPTERINFO));

      SearchList.lpAdapter[i+n]->Type = SERIAL_DS2480;
      SearchList.lpAdapter[i+n]->Num  = 0;

      /* Copy over device name */
      if (strlen(DevName[i]) < MAX_DEV_NAME_LEN)
         strcpy(SearchList.lpAdapter[i+n]->DevString, DevName[i]);
      else
         SearchList.lpAdapter[i+n]->DevString[0] = 0;
   }

   return;
}
/****************************************************************************
 *
 *   Releases API resources - 
 *       Should be called just before an application terminates.
 *
 */
void DLLEXPORT ReleaseCiBAPIResources()
{
  ReleaseCiBSearchList();
}
/****************************************************************************
 *
 *   Find all CiBs on all parallel ports
 *
 */
LPBYTE DLLEXPORT FindCiBs(LPBYTE lpCiBNum)
{
   BOOL AdapterFound;

   if (!SearchParamsSet)
   {
      /* Set default search parametere */
      SetCiBSearchParams(0, NULL);
   }

   *lpCiBNum = CiBsFound = SearchPorts(CiBROMs, 
                                       CiBPorts, 
                                       &AdapterFound,
                                       &SearchList);

   if (!AdapterFound)
   {
      SetCiBError(ERR_ADAPTER_NOT_FOUND);
   }
   else
   {
      if (!CiBsFound)
      {
         SetCiBError(ERR_NO_CIBS_FOUND);
      }
   }

   return CiBsFound ? CiBROMs : NULL;
}

/***************************************************************************
 *
 *   Selects the miac whose ROM matches the data pointed to by 
 * lpCiBRomPtr.
 *
 */
BOOL DLLEXPORT SelectCiB(LPBYTE lpCiBRomPtr)
{
   BOOL ROMIdFound = FALSE;
   BYTE n;
   WORD i;

   i = 0;
   do
   {
      if (!memcmp(lpCiBRomPtr, CiBROMs + i, 8))
      {
         /* Assign ROM data */
         memcpy(CurrCiB.ROMData, CiBROMs + i, 8);

         n = CiBPorts[i >> 3];

         /* Set current CiB structure info */
         CurrCiB.iBPort.Type = SearchList.lpAdapter[n]->Type;
         CurrCiB.iBPort.Num  = SearchList.lpAdapter[n]->Num;
         strcpy(CurrCiB.iBPort.DevString,
                SearchList.lpAdapter[n]->DevString);

         ROMIdFound = TRUE;
      }

      i += 8;
   }
   while (!ROMIdFound && (i < (CiBsFound << 3)));

   if (!ROMIdFound)
      SetCiBError(ERR_BAD_CIB_ROM);

   return ROMIdFound;
}

/***************************************************************************
 *
 *    Calls one of the "Comm" layer commands.  These are primarily used
 * as low level diagnostic tools.
 *
 */
BOOL DLLEXPORT SendCommCommand(BYTE CommCmd, 
                               LPCOMMPACKET lpCommCmd, 
                               LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = CommCmd;
   CP.GroupID = 0xFF;
   CP.Len     = 2 + lpCommCmd->Len;

   if (CP.Len)
      memcpy(CP.CmdData, lpCommCmd->CmdData, CP.Len);

   CmdRes = DispatchCommand(&CP, lpRP);

   /* Copy results to caller's LPCOMMPACKET structure */
   if (CmdRes)
   {
      lpCommCmd->Len = lpRP->DataLen;
      memcpy(lpCommCmd->CmdData, lpRP->CmdData, lpRP->DataLen);
   }

   return CmdRes;
}

/***************************************************************************
 *
 *   Sets a new common PIN for the selected CiB.
 *
 */
BOOL DLLEXPORT SetCommonPIN(LPCIBPIN lpCommonPIN, 
                            LPCIBPIN lpNewPIN, 
                            BYTE OptionByte, 
                            LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = SET_COMMON_PIN;
   CP.GroupID = 0;
   CP.Len     = lpCommonPIN->Len + lpNewPIN->Len + 5;

   /* Copy current common PIN into command buffer */
   CP.CmdData[0] = lpCommonPIN->Len;
   memcpy(&CP.CmdData[1], lpCommonPIN->PINData, lpCommonPIN->Len);

   /* Copy new common PIN into command buffer */
   CP.CmdData[1+lpCommonPIN->Len] = lpNewPIN->Len;
   memcpy(&CP.CmdData[2+lpCommonPIN->Len], lpNewPIN->PINData, lpNewPIN->Len);

   /* Set the common PIN option byte */
   CP.CmdData[CP.Len-3] = OptionByte;

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   If the supplied common PIN is correct, this command forces a master
 * erase (RAM clear) on the next power-up sequence..
 *
 */
BOOL DLLEXPORT MasterErase(LPCIBPIN lpCommonPIN, LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = MASTER_ERASE;
   CP.GroupID = 0;
   CP.Len     = lpCommonPIN->Len + 3;

   /* Copy current common PIN into command buffer */
   CP.CmdData[0] = lpCommonPIN->Len;
   memcpy(&CP.CmdData[1], lpCommonPIN->PINData, lpCommonPIN->Len);

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *  Create a new transaction group in the currently selected CiB.
 *
 */
BOOL DLLEXPORT CreateTransactionGroup(LPCIBPIN lpCommonPIN,
                                      LPCIBNAME lpGroupName,
                                      LPCIBPIN lpGroupPIN,
                                      BYTE GroupAttr,
                                      LPBYTE lpGroupID,
                                      LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   WORD      l; /* Index into CmdData array */

   /* Doesn't mean much now !  */
   if (GroupAttr)
      GroupAttr = 0;

   CP.CmdByte = CREATE_TRANSACTION_GROUP;
   CP.GroupID = 0;
   CP.Len     = lpCommonPIN->Len+lpGroupName->Len+lpGroupPIN->Len+5;

   /* Copy common PIN into command packet */
   l = CP.CmdData[0] = lpCommonPIN->Len;
   memcpy(&CP.CmdData[1], lpCommonPIN->PINData, l);

   /* Copy group name into command packet */
   CP.CmdData[++l] = lpGroupName->Len;
   memcpy(&CP.CmdData[++l], lpGroupName->NameData, lpGroupName->Len);
   l += lpGroupName->Len;

   /* Copy group PIN into command packet */
   CP.CmdData[l] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[++l], lpGroupPIN->PINData, lpGroupPIN->Len);

   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
      *lpGroupID = lpRP->GroupID;
   
   return CmdRes;
}

/***************************************************************************
 *
 *    Allows the caller to change the group PIN if current group PIN 
 * supplied is correct. 
 *
 */
BOOL DLLEXPORT SetGroupPIN(BYTE GroupID, 
                           LPCIBPIN lpGroupPIN, 
                           LPCIBPIN lpNewPIN, 
                           LPRETPACKET lpRP)
{
   CMDPACKET CP;
   WORD      l; /* Index into CmdData array */

   CP.CmdByte = SET_GROUP_PIN;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len+lpNewPIN->Len+4;

   /* Copy group PIN into command packet */
   l = CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, l);

   /* Copy new PIN into command packet */
   CP.CmdData[++l] = lpNewPIN->Len;
   memcpy(&CP.CmdData[++l], 
      (LPBYTE *) &lpNewPIN->PINData,
  lpNewPIN->Len);

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *  Create a new object in the transaction group specified by GroupID.
 *
 */
BOOL DLLEXPORT CreateCiBObject(BYTE GroupID, 
                               LPCIBPIN lpGroupPIN,
                               LPCIBOBJ lpNewObject,
                               LPBYTE lpObjectID,
                               LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   WORD      l; /* Index into CmdData array */

   CP.CmdByte = CREATE_OBJECT;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len+lpNewObject->Len+6;

   /* Copy group PIN into command packet */
   l = CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, l);

   memcpy(&CP.CmdData[++l], 
         (LPBYTE *) &lpNewObject->Attr,
         lpNewObject->Len + 3);

   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
      *lpObjectID = lpRP->CmdData[0];
   
   return CmdRes;
}

/***************************************************************************
 *
 *   Reads the name of the transaction group specified by GroupID.
 *
 */
BOOL DLLEXPORT ReadGroupName(BYTE GroupID, 
                             LPCIBNAME lpGroupName, 
                             LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   
   CP.CmdByte = READ_GROUP_NAME;
   CP.GroupID = GroupID;
   CP.Len     = 2;
   
   CmdRes = DispatchCommand(&CP, lpRP);
   
   /* Copy group name from return packet to group name buffer */
   if (CmdRes)
   {

      if (lpRP->CmdData[0] > '0')// why not >= '0'
      { 
         lpGroupName->Len = lpRP->DataLen;
         memcpy(lpGroupName->NameData, lpRP->CmdData, lpGroupName->Len);
      }
      else
      { 
         // Work around length bug in earlier firmware versions
         //lpGroupName->Len = lpRP->DataLen - 1;
         lpGroupName->Len = (lpRP->DataLen)?(lpRP->DataLen - 1):0;
         memcpy(lpGroupName->NameData, &lpRP->CmdData[1], lpGroupName->Len);
      }
   }
   
   return CmdRes;
}

/***************************************************************************
 *
 *   Checks the CRC-16 of the group specified by GroupID.
 *
 */
BOOL DLLEXPORT CheckGroupCRC(BYTE GroupID, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   
   CP.CmdByte = CHECK_GROUP_CRC;
   CP.GroupID = GroupID;
   CP.Len     = 2;
   
   CmdRes = DispatchCommand(&CP, lpRP);
   
   return CmdRes;
}

/***************************************************************************
 *
 *    Reads the object specified by ObjectID in the transaction group
 * specified by GroupID.
 *
 */
BOOL DLLEXPORT ReadCiBObject(BYTE GroupID, 
                             LPCIBPIN lpGroupPIN, 
                             BYTE ObjectID,
                             LPCIBOBJ lpObject,
                             LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;
   
   CP.CmdByte = READ_OBJECT;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;
   
   /* Copy group PIN into command packet */
   l = CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, l);
   
   CP.CmdData[l+1] = ObjectID;
   
   CmdRes = DispatchCommand(&CP, lpRP);
   
   if (CmdRes)
   {
      /* Copy object data returned by the CiB */
      memcpy((LPBYTE *) lpObject, 
              &lpRP->CmdData[1], 
              lpRP->DataLen - 1);
   }
   
   return CmdRes;
}

/***************************************************************************
 *
 *    Writes the object specified by ObjectID in the transaction group
 * specified by GroupID.
 *
 */
BOOL DLLEXPORT WriteCiBObject(BYTE GroupID, 
                              LPCIBPIN lpGroupPIN, 
                              BYTE ObjectID,
                              LPCIBOBJ lpObject,
                              LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BYTE      l;
   
   CP.CmdByte = WRITE_OBJECT;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + lpObject->Len + 5;
   
   /* Copy group PIN into command packet */
   l = CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, l);
   
   CP.CmdData[++l] = ObjectID;
   CP.CmdData[++l] = lpObject->Len;
   memcpy(&CP.CmdData[++l], lpObject->ObjData, lpObject->Len);
   
   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   Locks the transaction group specified by GroupID.
 *
 */
BOOL DLLEXPORT LockGroup(BYTE GroupID, LPCIBPIN lpGroupPIN, LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = LOCK_GROUP;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 3;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   Deletes the transaction group specified by GroupID.
 *
 */
BOOL DLLEXPORT DeleteGroup(BYTE GroupID, 
                           LPCIBPIN lpGroupPIN, 
                           LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = DELETE_GROUP;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 3;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);

   return DispatchCommand(&CP, lpRP);
}


/***************************************************************************
 *
 *   Uses the supplied group name to retrieve the group's ID.
 *
 */
BOOL DLLEXPORT GetGroupID(LPCIBNAME lpGroupName,
                          LPBYTE lpGroupID, 
                          LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = GET_GROUP_ID;
   CP.GroupID = 0;
   CP.Len     = lpGroupName->Len+3;
   
   /* Copy group PIN into command packet */
       CP.CmdData[0] = lpGroupName->Len;
       memcpy(&CP.CmdData[1], lpGroupName->NameData, CP.CmdData[0]);
   
   CmdRes = DispatchCommand(&CP, lpRP);
   
   if (CmdRes)
      *lpGroupID = lpRP->GroupID;
   
   return CmdRes;
}

/***************************************************************************
 *
 *   Locks the CiB and returns the audit trail size.
 *
 */
BOOL DLLEXPORT LockCiB(LPCIBPIN lpCommonPIN, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = LOCK_CIB;
   CP.GroupID = 0;
   CP.Len     = lpCommonPIN->Len + 3;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpCommonPIN->Len;
   memcpy(&CP.CmdData[1], lpCommonPIN->PINData, lpCommonPIN->Len);

   CmdRes = DispatchCommand(&CP, lpRP);

   return CmdRes;
}

/***************************************************************************
 *
 *    Returns the configuration structure of the currently selected CiB.
 *
 */
BOOL DLLEXPORT GetCiBConfiguration(LPCIBINFO lpConfig, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = READ_CONFIG_INFO;
   CP.GroupID = 0;
   CP.Len     = 2;

   CmdRes = DispatchCommand(&CP, lpRP);

   /* Copy version string to caller's buffer */
   if (CmdRes)
      memcpy(&lpConfig->GroupNum,lpRP->CmdData, sizeof(CIBINFO));

   return CmdRes;
}

/***************************************************************************
 *
 *    Returns the MIAC_ST byte which includes any error bits.
 *
 */
BOOL DLLEXPORT GetCiBStatusByte(LPBYTE lpCiBStatus, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = GET_STATUS_BYTE;
   CP.GroupID = 0;
   CP.Len     = 2;

   CmdRes = DispatchCommand(&CP, lpRP);

   /* Copy version string to caller's buffer */
   if (CmdRes)
      *lpCiBStatus = lpRP->CmdData[0];
   else
      *lpCiBStatus = 0;

   return CmdRes;
}

/***************************************************************************
 *
 *    Locks the object specified by ObjectID in the transaction group
 * specified by GroupID.
 *
 */
BOOL DLLEXPORT SetCiBObjectAttr(BYTE GroupID, 
                                LPCIBPIN lpGroupPIN,
                                BYTE ObjectID, 
                                BYTE Attr,
                                LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BYTE      l;
   
   switch (Attr)
   {
      case LOCKED:
         CP.CmdByte = LOCK_OBJECT;
      break;
   
      case PRIVATE:
         CP.CmdByte = PRIVATIZE_OBJECT;
      break;
   
      case DESTRUCT:
         CP.CmdByte = DESTRUCTIBLE_OBJECT;
      break;
   
      default:
      CP.CmdByte = 0;
   }
   
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   /* Copy group PIN into command packet */
   l = CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, l);

   CP.CmdData[++l] = ObjectID;

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *    Reads the RAM remaining (unused by groups) in the currently selected
 * CiB.
 *
 */
BOOL DLLEXPORT ReadFreeRAM(LPWORD lpFreeRAM, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = READ_FREE_RAM;
   CP.GroupID = 0;
   CP.Len     = 2;

   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   memcpy((BYTE *) lpFreeRAM, lpRP->CmdData, 2);

   return CmdRes;
}

/***************************************************************************
 *
 *   Invokes the random number generator self-tests. These tests are
 * described in FIPS 140-1.
 *
 */
BOOL DLLEXPORT TestRNG(LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = TEST_RNG;
   CP.GroupID = 0;
   CP.Len     = 2;

   ExecTimeMS = 4000;

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   Reads the real time clock of the selected CiB and returns the 4 MSBs
 * in the long word pointed to by lpRTCSeconds.
 *
 */
BOOL DLLEXPORT ReadRealTimeClock(LPDWORD lpRTCSeconds, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = READ_REAL_TIME_CLOCK;
   CP.GroupID = 0;
   CP.Len     = 2;

   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   memcpy((BYTE *) lpRTCSeconds, lpRP->CmdData, 4);

   return CmdRes;
}

/***************************************************************************
 *
 *    Reads the real time clock and adds that value to the clock offset
 * specified by ObjectID.
 *
 */
BOOL DLLEXPORT ReadTrueTimeClock(BYTE GroupID,
                                 LPCIBPIN lpGroupPIN,
                                 BYTE ObjectID,
                                 LPDWORD lpSeconds,
                                 LPRETPACKET lpRP)
{
   BOOL      CmdRes;  
   CMDPACKET CP;

   CP.CmdByte = READ_TRUE_TIME_CLOCK;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);

   CP.CmdData[lpGroupPIN->Len + 1] = ObjectID;

   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
      memcpy((BYTE *) lpSeconds, lpRP->CmdData, 4);

   return CmdRes;
}

/***************************************************************************
 *
 *  Invoke script specified by Group ID, Object ID
 *
 */
BOOL DLLEXPORT InvokeScript(BYTE GroupID, 
                            LPCIBPIN lpGroupPIN, 
                            BYTE ObjectID, 
                            WORD RunMS,
                            LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = INVOKE_SCRIPT;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);

   CP.CmdData[lpGroupPIN->Len + 1] = ObjectID;

   /* Set initial guess of script execution time (in milliseconds) */
   ExecTimeMS = RunMS;

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   Change the name of the transaction group specified by GroupID.
 *
 */
BOOL DLLEXPORT ChangeGroupName(BYTE GroupID, 
                               LPCIBPIN lpGroupPIN, 
                               LPCIBNAME lpGroupName, 
                               LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BYTE      l;

   CP.CmdByte = CHANGE_GROUP_NAME;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + lpGroupName->Len + 4;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy new group name into command buffer */
   CP.CmdData[l++] = lpGroupName->Len;
   memcpy(&CP.CmdData[l], lpGroupName->NameData, lpGroupName->Len);

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *   Retrieve the firmware version string from the currently selected
 * CiB.
 *
 */
BOOL DLLEXPORT ReadFirmwareVersionID(LPCIBNAME lpVerString, LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;

   CP.CmdByte = GET_VER_STRING;
   CP.GroupID = 0;
   CP.Len     = 2;

   CmdRes = DispatchCommand(&CP, lpRP);

   /* Copy version string to caller's buffer */
   if (CmdRes)
   {
      lpVerString->Len = lpRP->DataLen;
      memcpy(lpVerString->NameData, lpRP->CmdData, lpRP->DataLen);
   }

   return CmdRes;
}

/***************************************************************************
 *
 *   Requests nBytes bytes of random data from the currently selected
 * CiB
 *
 */
BOOL DLLEXPORT ReadRandomBytes(BYTE nBytes, 
                               LPBYTE lpRandomBuffer, 
                               LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;   

   CP.CmdByte = GET_RANDOM_DATA;
   CP.GroupID = 0;
   CP.Len     = 3;

   /* Requested number of random bytes  */
   CP.CmdData[0] = nBytes;

   /* This takes a while */
   ExecTimeMS = 250;

   CmdRes = DispatchCommand(&CP, lpRP);

   /* Copy over random data buffer */
   if (CmdRes)
     memcpy(lpRandomBuffer, lpRP->CmdData, lpRP->DataLen);

   return CmdRes;
}

/***************************************************************************
 *
 *    Generates an entire RSA key set and returns the object IDs of the 
 * newly created elements.
 *
 */
BOOL DLLEXPORT GenerateRSAKeySet(BYTE GroupID,
                                 LPCIBPIN lpGroupPIN,
                                 BYTE ModulusSize,
                                 LPBYTE lpModulusID,
                                 LPBYTE lpPublicExpID,
                                 LPBYTE lpPrivateExpID,
                                 LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;

   CP.CmdByte = GEN_RSA_KS;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy number of bytes in modulus */
   CP.CmdData[l] = ModulusSize;

   ExecTimeMS = 4000;
   
   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   {
      *lpModulusID    = lpRP->CmdData[0];
      *lpPublicExpID  = lpRP->CmdData[1];
      *lpPrivateExpID = lpRP->CmdData[2];
   }

   return CmdRes;
}

/***************************************************************************
 *
 *    Generates an entire RSA key set and returns the object IDs of the 
 * newly created elements.
 *
 */
BOOL DLLEXPORT GenerateRSAModAndExp(BYTE GroupID,
                                    LPCIBPIN lpGroupPIN,
                                    BYTE ModulusSize,
                                    BYTE PublicExpID,
                                    LPBYTE lpModulusID,
                                    LPBYTE lpPrivateExpID,
                                    LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;

   CP.CmdByte = GEN_RSA_MOD_AND_EXP;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 5;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy number of bytes in modulus */
   CP.CmdData[l] = ModulusSize;
   CP.CmdData[l+1] = PublicExpID;

   ExecTimeMS = 4000;
   
   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   {
      *lpModulusID    = lpRP->CmdData[0];
      *lpPrivateExpID = lpRP->CmdData[1];
   }

   return CmdRes;
}

/***************************************************************************
 *
 *    Generates an entire RSA key set and returns the object IDs of the 
 * newly created elements. This command differs from GenerateRSAKeySet in 
 * 2 ways. Neither exponent is made private and none of the objects created
 * by this command will have the "Cib Created" attribute bit set in the
 * object attribute byte.
 *
 */
BOOL DLLEXPORT GenerateRSAKeySetNP(BYTE GroupID,
                                   LPCIBPIN lpGroupPIN,
                                   BYTE ModulusSize,
                                   LPBYTE lpModulusID,
                                   LPBYTE lpPublicExpID,
                                   LPBYTE lpPrivateExpID,
                                   LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;

   CP.CmdByte = GEN_RSA_KS_NO_PRIVATE;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy number of bytes in modulus */
   CP.CmdData[l] = ModulusSize;

   ExecTimeMS = 4000;
   
   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   {
      *lpModulusID    = lpRP->CmdData[0];
      *lpPublicExpID  = lpRP->CmdData[1];
      *lpPrivateExpID = lpRP->CmdData[2];
   }

   return CmdRes;
}

/***************************************************************************
 *
 *    Generates a random private exponent
 */
BOOL DLLEXPORT GenerateRandomExponent(BYTE GroupID,
                                      LPCIBPIN lpGroupPIN,
                                      BYTE ExponentSize,
                                      LPBYTE lpPrivateExpID,
                                      LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;

   CP.CmdByte = GEN_RANDOM_EXP;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy number of bytes in modulus */
   CP.CmdData[l] = ExponentSize;

   ExecTimeMS = 250;
   
   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
      *lpPrivateExpID = lpRP->CmdData[0];

   return CmdRes;
}

/***************************************************************************
 *
 *    Turn off RSA key set generation capability.
 *
 */
BOOL DLLEXPORT DisableKeySetGeneration(LPCIBPIN lpCommonPIN, LPRETPACKET lpRP)
{
   CMDPACKET CP;

   CP.CmdByte = DISABLE_KS_GEN;
   CP.GroupID = 0;
   CP.Len     = lpCommonPIN->Len + 3;

   /* Copy current common PIN into command buffer */
   CP.CmdData[0] = lpCommonPIN->Len;
   memcpy(&CP.CmdData[1], lpCommonPIN->PINData, lpCommonPIN->Len);

   ExecTimeMS = 4000;

   return DispatchCommand(&CP, lpRP);
}

/***************************************************************************
 *
 *    Generate prime number. 
 *
 */
BOOL DLLEXPORT GeneratePrime(BYTE GroupID,
                             LPCIBPIN lpGroupPIN, 
                             BYTE PrimeSize, 
                             LPBIGNUM lpPrime, 
                             LPRETPACKET lpRP)
{
   CMDPACKET CP;
   BOOL      CmdRes;
   BYTE      l;

   CP.CmdByte = GEN_PRIME;
   CP.GroupID = GroupID;
   CP.Len     = lpGroupPIN->Len + 4;

   l = lpGroupPIN->Len + 1;

   /* Copy group PIN into command buffer */
   CP.CmdData[0] = lpGroupPIN->Len;
   memcpy(&CP.CmdData[1], lpGroupPIN->PINData, lpGroupPIN->Len);
   /* Copy number of bytes in modulus */
   CP.CmdData[l] = PrimeSize;
   
   CmdRes = DispatchCommand(&CP, lpRP);

   if (CmdRes)
   {
      /* Copy prime number into BigNum */
      lpPrime->Len = PrimeSize;
      memcpy(lpPrime->NumArr, lpRP->CmdData, lpPrime->Len);
   }

   return CmdRes;
}

/***************************************************************************
 *
 *    Set new error code.
 *
 */
void SetCiBError(WORD ErrCode)
{
   LAST_ERROR = ErrCode;
}

/***************************************************************************
 *
 *    Return the last error.
 *
 */
WORD DLLEXPORT GetCiBError()
{
   return LAST_ERROR;
}

/***************************************************************************
 *
 *    Send command to the device via the "COMM LEVEL" DLL. This funcion
 * sets LAST_ERROR in the event of any communication error.
 *
 */
BOOL DispatchCommand(LPCMDPACKET lpCP, LPRETPACKET lpRP)
{
   WORD RLen;
   WORD ErrCode;
   WORD Delay = 0;

   if (ExecTimeMS || MinTimeMS)
   {
      Delay = (ExecTimeMS + MinTimeMS) / 250;   

      if (Delay > 0x0F)
         Delay = 0x0F;
   }

   /* Send command to the device */
   ErrCode = SendCiBMessage((LPBYTE) &lpCP->Len, 
                             (WORD) (lpCP->Len+1), /* Include length */
                             Delay, 
                             &CurrCiB);

   if (!ErrCode)
   {
      RLen = MAX_BUF_SIZE;
      /* Read the devices response */
      ErrCode = RecvCiBMessage((LPBYTE) &lpRP->CSB, 
                                &RLen, 
                                &CurrCiB);
   }

   switch (ErrCode)
   {
      case 0:
         /* All comm level stuff went well, so error code is set to CSB */
         SetCiBError(lpRP->CSB);
      break;

      default:
         SetCiBError(ERR_COMM_FAILURE);
   }

   /* Reset execution time */
   ExecTimeMS = 0;

   return LAST_ERROR ? FALSE : TRUE;
}

void DLLEXPORT SetMinRun(WORD rt)
{
   MinTimeMS = rt;
}

