/****************************************************************************
 *
 *       ds1996demo.c               
 *
 *     This program prints the ROM data of all Buttons found on the serial
 * port, and reads the secure memory of any attached ds1992s, 3s, 4s or 6s. 
 * After all the memory in the iButton has been displayed each byte of each 
 * page will be overwitten with the character received from the command line. 
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>

#include "sauthunx.h"
#include "ds1996demo.h"

void main(int argc, char *argv[])
{
   SMEMDATA iBMemRead;
   SPDATA   iBScratchWr;
   SPCOPY   iBScratchCopy;
   uchar    OutByte, ch;
   uchar    *addr;  
   ushort   PageNum, i, j;
   char     DeviceName[] = "/dev/ttya";

   if (argc == 2)
   {
      OutByte = argv[1][0];
      /* Tell access system which adapter type to expect */
      SetAdapterType(SERIAL_DS2480, DeviceName);
   
      /* See if device driver is installed */
      if (iBDOWCheck())                     
      {
         iBSetup(1);
         iBKeyOpen();   
         SetAdapterSpeed(9600);
   
         /* Look for first part on bus */
         if (iBFirst())                   
         {
            do 
            {
               printf("\nRom Data: ");         
               /* Get pointer to ROM data */
               addr = iBROMData();           
               for (i = 0; i < 8; i++)    
                  printf("%2.2X ", addr[i]);
               printf("\n");
   
               PageNum = 0;
               /* 
                * Check family code to see if this is a button we're   
                * interested in. If so set correct number of pages.
                */ 
               switch (addr[0])
               {
                  case DS1992_FC:
                     printf("\nReading the memory of a DS1992\n");
                     PageNum = DS1992_PAGE_NUM;
                  break;

                  case DS1993_FC:
                     printf("\nReading the memory of a DS1993\n");
                     PageNum = DS1993_PAGE_NUM;
                  break;

                  case DS1994_FC:
                     printf("\nReading the memory of a DS1994\n");
                     PageNum = DS1994_PAGE_NUM;
                  break;
 
                  case DS1996_FC:
                     printf("\nReading the memory of a DS1996\n");
                     PageNum = DS1996_PAGE_NUM;

                     /*
                      * Since the DS1996s support overdrive put the
                      * 1-wire bus in overdrive and increase the serial
                      * port data rate (The actual DS2480 supports higher
                      * data rates).
                      */
                     iBOverdriveOn();
                     /* 
                      * 57600 and 115200 would be better but these data rates 
                      * are unsupported on some platforms. The DS2480 however 
                      * should have no trouble with these higher data rates.
                      */
                     SetAdapterSpeed(19200);
               }

               if (PageNum)
               {
                  /* Read all the secure memory pages 1st */
                  if (iBFastAccess())
                  {
                     iBMemRead.CmdByte = READ_SECURE_MEMORY;
                     /* Begin reading at memory address 0 */
                     iBMemRead.TargetAddrLow = iBMemRead.TargetAddrHigh = 0;
                     /* Send 0xFFs for each byte read */
                     memset(iBMemRead.SMemBuffer, 0xFF, PageNum * PAGE_LEN);
                  
                     iBDataBlock(&iBMemRead.CmdByte, PageNum * PAGE_LEN + 3);

                     /* Display received data on page boundaries */
                     for (i = 0; i < PageNum; i++)
                     {
                        printf("\nPage #%d:", i);
                        for (j = 0; j < PAGE_LEN; j++)
                        {
                           ch = iBMemRead.SMemBuffer[i*PAGE_LEN + j]; 
                           if (ch < 0x20)
                              putchar('.');
                           else
                              putchar(ch);
                        }
                     }

                     printf("\n");

                     /* Reprogram all memory pages */
                     for (i = 0; i < PageNum; i++)
                     {
                        if (iBFastAccess())
                        {
                           printf("Writing scratchpad\n");
   
                           /* Set write scratchpad command byte */
                           iBScratchWr.CmdByte = WRITE_SCRATCHPAD;
                           /* Set the low byte of the target address */
                           iBScratchWr.TargetAddrLow  = i << 5;
                           /* Set the high byte of the target address */
                           iBScratchWr.TargetAddrHigh = i >> 3;
                           /* Initialize scratchpad buffer */
                           memset(iBScratchWr.SPBuffer, OutByte, PAGE_LEN);
                           
                           iBDataBlock(&iBScratchWr.CmdByte, PAGE_LEN+3);

                           if (iBFastAccess())
                           {
                              /*
                               * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                               *    This code is assuming that the iButton(s) 
                               * are dwelled. In a touch environment the 
                               * scratchpad should be read back and confirmed 
                               * before proceeding with the copy scratchpad 
                               * command.
                               * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                               */
                              printf("Copying scratchpad to page #%d\n", i);
      
                              /* Set copy scratchpad command byte */
                              iBScratchCopy.CmdByte = COPY_SCRATCHPAD;
                              /* Set the low byte of the target address */
                              iBScratchCopy.TargetAddrLow  = i << 5;
                              /* Set the high byte of the target address */
                              iBScratchCopy.TargetAddrHigh = i >> 3;
                              /* Copy entire scratchpad to specified page */
                              iBScratchCopy.Offset = 0x1F; 

                              iBDataBlock(&iBScratchCopy.CmdByte, 
                                          sizeof(SPCOPY));
                           }
                           else
                              printf("\n!!!Unable to access iButton\n");
                        }
                        else
                           printf("\n!!!Unable to access iButton\n");
                     }
                  }
               }

               /* If we put the bus in overdrive restore regular speed */
               if (addr[0] == DS1996_FC)
               {
                  iBOverdriveOff();
                  SetAdapterSpeed(9600);
               }
            }
            while (iBNext());  /* Look at all parts on port */

            printf("\n");
         }
         else 
            printf("No buttons found on port\n");
   
         /* Release control of port */
         iBKeyClose();                                
      }
      else
         printf("DOW device driver not installed.\n");
   }
   else
      printf("\nUsage:ds1996demo out_char\n");
}


